from scipy.interpolate import CubicSpline
from scipy.signal import savgol_filter
import numpy as np
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
import matplotlib.image as mpimg
import matplotlib as mpl
tex_settings = {                      
     "text.usetex": True,
     "font.family": "sans-serif",
     "font.serif" : "Computer Modern Roman",
     "text.latex.preamble": [
         r"\usepackage{siunitx}",
         r" \usepackage[version=4]{mhchem}",
         r"\sisetup{detect-all}"
         ]
     }
mpl.rcParams.update(tex_settings)
import sys
sys.path.append('../../numerical_analysis/')
import xarray as xr
from pathlib import Path

# load in-plane calculation data
MAIN_FOLDER = Path('../../data/theory/2D_boundary')
data = xr.load_dataset(MAIN_FOLDER/Path('2021-03-16-19-43-00'))
data = data['__xarray_dataarray_variable__']
B = data['By']
rho = 1/data[:, 0, 0, 0]
rho = (rho-rho[0])/rho[0]
ddrho = np.gradient(np.gradient(rho))

# load experimental data from Carsten
file_experiments = ['Fig1_B_PdCoO2.dat', 'Fig1_B_PtCoO2_ac1_3p9um.dat', 'Fig1_B_PtCoO2_ac2_4p8um.dat']
file_experiments = ['../../data/experiment/Carsten_data/Figures/Fig1/'+i for i in file_experiments]

def exp_interpolation(file_exp):
    exp_data = np.loadtxt(file_exp, skiprows=3)
    exp_B = np.flip(exp_data[:,0])
    exp_rho = np.flip(exp_data[:,-1])
    del_index = np.array([True])
    while not del_index.size == 0:
        test = np.diff(exp_B)
        del_index = np.where(test<=0)[0]
        exp_B = np.delete(exp_B, del_index)
        exp_rho = np.delete(exp_rho, del_index)
    exp_rho_fn = CubicSpline(exp_B,exp_rho)
    return exp_rho_fn

# Find the derivative under smoothing for Pd
exp_ddrho = savgol_filter(exp_interpolation(file_experiments[0])(B),27,3,deriv=2)

column_size = 5.90666
fig = plt.figure(figsize=(column_size,column_size/1.61803398875))
plot1 = plt.subplot2grid((2, 2), (0, 0), colspan=2, fig=fig)
plot2 = plt.subplot2grid((2, 2), (1, 0), fig=fig)
plot3 = plt.subplot2grid((2, 2), (1, 1), fig=fig)

x = np.arange(1, 10)

set_up = mpimg.imread('../inkscape/sample_geometry.png')
plot1.imshow(set_up)
plot1.axis("off")

#plot2.plot(x, x**0.5)
plot2.plot(B, exp_interpolation(file_experiments[0])(B), c='C0')
plot2.plot(B, rho,label='Semiclassical theory', c='C0',ls='--')

ticks = np.arange(0, 12, 2)

plot2.set_ylim([0,45])
plot2.set_ylabel('$ \\Delta \\rho(B)/\\rho(0) $',fontsize = 14)
plot2.set_xlim([0,10])
plot2.set_xlabel('$ B[T] $',fontsize=14)
plot2.set_xticks(ticks)

# Plot 2nd order magnetoresistance
plot3.plot(B,exp_ddrho/exp_interpolation(file_experiments[0])(B), c='C0')
plot3.plot(B,ddrho/rho, ls='--', c='C0')
plot3.set_ylabel('$ (d^2 \\rho / dB^2) / \\rho $',fontsize=14)
plot3.set_xlim([0,10])
plot3.set_ylim([-0.0001,0.0002])
plot3.set_xlabel('$ B[T] $',fontsize=14)
plot3.set_yticklabels([])
plot3.set_xticks(ticks)


lines = [Line2D([0], [0], c='C0' ), Line2D([0], [0], ls='--', c='C0' )]
labels = ['Experiment', 'Theory']
plt.rcParams['legend.title_fontsize'] = 9
leg = plot2.legend(lines, labels, title=r"\ce{PdCoO2} $L=$ \SI{3.9}{\micro\meter}, $W=$ \SI{5.5}{\micro\meter}:", frameon=False, fontsize=9)
leg._legend_box.align = "left"
plt.tight_layout()
plt.savefig("../figure1.svg",bbox_inches='tight',transparent=False, dpi=1000)
