import numpy as np
from numpy import pi
from scipy import constants
from matplotlib.gridspec import GridSpec
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import sys
sys.path.append('../../numerical_analysis/')
from modules.quasi2d import Quasi2D
from modules.Pd_dispersion import Pd_energy
import matplotlib as mpl
tex_settings = {                      
     "text.usetex": True,
     "font.family": "sans-serif",
     "font.serif" : "Computer Modern Roman"
     }
mpl.rcParams.update(tex_settings)
L_B_SI = np.sqrt(constants.hbar/(constants.elementary_charge))

################################## Left Column Plot - 2D trajectory plot ##################################

# Define different trajectories via their phase
phi = np.linspace(0,2*np.pi,12, endpoint=False)
x_2d = np.linspace(0,1,1000)
def curve(x,freq=1,phi=0):
    return np.cos(2*np.pi*freq*x+phi)
    
# Define trajectories with random scattering events
def scattering_curve(x, scatt,freq=2, phi = 0):
    y = np.zeros(x.shape[0])
    nodes = np.nonzero(np.random.poisson(lam=scatt, size=x.shape[0]))[0]
    if nodes.shape[0] == 0:
        nodes = np.array([x.shape[0]-1])
    else:
        nodes[-1] = x.shape[0]-1
    
    phases = np.random.uniform(size=nodes.shape[0])*2*np.pi
    phases[0] = phi
    
    y_0 = 0
    x_ind = 0
    for index, i in enumerate(nodes):
        if i == 0:
            continue
        else:
            y_temp = curve(x[x_ind:i+1],phi=phases[index],freq=freq)
            y_temp = y_temp - y_temp[0]
            y[x_ind:i+1] = y_temp + y_0
            y_0 = y[i]
            x_ind = i+1
    return y

def add_symbol(ax,text,coords=(-0.15,0.85)):
    ax.text(*coords, text, transform=ax.transAxes, size=18, weight='bold')

# Create the overall figure template
np.random.seed(12)
text_width = 5.90666
figr, axs = plt.subplots(ncols=2, nrows=3, figsize=(text_width,text_width/1.61803398875),sharex=True)
gs = axs[0, 1].get_gridspec()

# Remove the underlying axes
for axi in axs[0:, -1]:
    axi.remove()
    
# Define the subplots on the left column of the figure
axA = axs[0][0] 
axB = axs[1][0] 
axC = axs[2][0] 
ax_column1 = [axA,axB,axC]
labels = ['(A)','(B)','(C)']

# Define the larger subplot on the right column
axbig = figr.add_subplot(gs[0:, -1])

# Plot the three subplots on the left column
for i in phi:
    axA.plot(x_2d,curve(x_2d,freq=3,phi=i)-curve(x_2d[0],freq=3,phi=i),alpha=0.7)
    axB.plot(x_2d,curve(x_2d,freq=3.5,phi=i)-curve(0,freq=3.5,phi=i),alpha=0.7)
    axC.plot(x_2d, scattering_curve(x_2d,0.0015,freq=3,phi=i),alpha=0.7)
for count,i in enumerate(ax_column1):
    i.spines['top'].set_visible(False)
    i.spines['bottom'].set_visible(False)
    i.axhline(0, ls='--',c='black')
    i.set_ylim([-2,2])
    i.set_xlim([0,1])
    i.set_yticks([])
    i.set_xticks([])
    add_symbol(i,labels[count])
axC.set_xlabel('$x$', )
axB.set_ylabel('$z$',rotation=0,labelpad=10)

################################## Right Column Plot - Out-of-plane field trajectories ##################################
  
# Obtain electron trajectories
delafossite = Quasi2D(Pd_energy, N=1000, FS_ROTATE = 0)

@np.vectorize
def k_fermi(theta):
    if theta>3*pi/2:
        theta = theta-2*pi
    return delafossite._kf(theta)*1e-10/L_B_SI

def fs(theta=np.linspace(0,2*np.pi,100),rotate=0):
    theta = theta-rotate
    r = np.array(( (np.cos(rotate), -np.sin(rotate)),
               (np.sin(rotate),  np.cos(rotate)) ))
    result = k_fermi(theta)*np.array((np.cos(theta),np.sin(theta)))
    return r.dot(result)

def g_trajectory(t1,t2):
    x, y = fs(theta=np.linspace(t1,t2,500),rotate=pi/6)
    x0 = np.min(x)
    x = x-x0
    y = y - np.min(y)
    return x,y

# Add oscillatory behaviour to the trajectories   
def wavy_noise(x,freq = 100, A = 0.05):
    return A*np.sin(freq*x) 

# Make a cavalier transformation for an isometric plot
angle = np.pi/180*45+pi
cavalier_T = np.array(([1,0,0.5*np.cos(angle)],[0,1,0.5*np.sin(angle)],[0,0,0]))

# Set up the axes of the box
z = np.array([0,0,0.2])
x = np.array([1,0,0])
y = np.array([0,1,0])
zero = np.array([0,0,0])  
shift = np.array([-0.1,-0.06,0])

# Define the required Cavalier transformations for the isometric plot
def xy_T(plane_vectors):
    transformed_plane = np.dot(cavalier_T,plane_vectors.T)
    return transformed_plane[0],transformed_plane[1]

def arrow(begin,end,ax):
    ax.annotate("",
        xy = end, xycoords='data',
        xytext=begin,textcoords='data',
        arrowprops=dict(arrowstyle="->", connectionstyle="arc3", lw = 1.5),
                )

def text_ann(string, pos, ax):
    ax.annotate(string,xy=(0,0),xytext=pos,size=10)
    
def coord_sys(**kwargs):
    # Plot z
    arrow(xy_T(z+np.array([0.0,-0.01,0])),xy_T(z+y),**kwargs)
    text_ann("$Y$",xy_T(z+y+shift),**kwargs)
    arrow(xy_T(z+np.array([-0.005,-0.005,0])),xy_T(zero),**kwargs)
    text_ann("$Z$",xy_T(zero+np.array([-0.06,0, 0])),**kwargs)
    arrow(xy_T(z+np.array([-0.0125,0,0])),xy_T(z+x),**kwargs)
    text_ann("$X$",xy_T(z+x+shift),**kwargs)

def sample_cavalier(ax):
    coord_sys(ax=ax)
    
    xy = np.array([[0,0,0],y,y+x,x])
    xz = np.array([[0,0,0],z,z+x,x])
    yz = np.array([[0,0,0],z,z+y,y])
    
    ax.fill(*xy_T(xy), alpha=0.2, edgecolor=None,facecolor='b')
    ax.fill(*xy_T(xz), alpha=0.5, edgecolor=None,facecolor='b')
    ax.fill(*xy_T(yz), alpha=0.5, edgecolor=None,facecolor='b')
    
    ax.fill(*xy_T(xy+z), alpha=0.05, edgecolor=None,facecolor='b')
    ax.fill(*xy_T(xz+y), alpha=0.25, edgecolor=None,facecolor='b')
    ax.fill(*xy_T(yz+x), alpha=0.25, edgecolor=None,facecolor='b')
    
    ax.set_xlim([-0.5,2])
    ax.set_ylim([-0.5,2])

# Compute the trajectories
traj1 = g_trajectory(pi+pi/4,2*pi)
norm = np.max(traj1[0])
traj1 = traj1/norm
traj1[1] += 1/2-traj1[1][0]
z_shift = z[2]/2 - wavy_noise(traj1[0],freq=200, A = 0.025)
def traj_cavalier(traj,z = 0.5):
    return np.dot(cavalier_T,np.array([traj[0],traj[1],z]))[:-1]
def traj_generator(t1,t2,z = 0.1,wiggle=True):
    traj = g_trajectory(t1,t2)/norm
    traj[1] += 1/2-traj[1][0]
    if wiggle == True:
        z += wavy_noise(traj[0],freq=200, A = 0.05)
    return traj_cavalier(traj,z)

# Plot the 3D isometric plot on the right column of the figure
def isometric_plot(ax):
    sample_cavalier(ax=ax)
    ax.plot(*traj_cavalier(traj1,z_shift),color='b',zorder=0)
    ax.plot(*traj_generator(pi+pi/7,-pi/7+2*pi),color='b',zorder=0)
    ax.plot(*traj_generator(-pi/7,pi/7),color='r',zorder=0)
    ax.plot(*traj_generator(-pi/10,pi/10),color='r',zorder=0)
    ax.set_xlim([-0.1,1.1])
    ax.set_ylim([-0.1,1.0])
    ax.set_axis_off()
isometric_plot(axbig)
add_symbol(axbig,'(D)',coords=(-0.1,0.96))
plt.savefig('../figure2.svg',bbox_inches='tight',transparent=False, dpi=1000)
