import numpy as np
import xarray as xr
from scipy.constants import constants
from pathlib import Path
from matplotlib import pyplot as plt
from matplotlib import rc
rc('font',**{'family':'serif','serif':['Computer Modern Roman']})
rc('text', usetex=True)

# find the nearest index near a value
def index_find(array, value):
    return int(np.abs(array-value).argmin())

# load experimental data
files = list(Path('../../data/experiment/Carsten_data/Figures/Fig3/').glob('*deg.dat'))
files.sort(key=lambda s: int(str(s).split('rotac_')[1].split('deg')[0]))

# choose the correct angles for the plot (0 - 45 degrees)
ang_start = 12
ang_end = 22
files = files[ang_start:ang_end]
N_exp = ang_end-ang_start
shift = 1

# compute the variance of the 0 degree tilt data from 1 to 3 Tesla
data_in = np.loadtxt(files[-1],skiprows=2)
B_exp = data_in[:,0]
ind_B_st = index_find(B_exp,1)
ind_B_end = index_find(B_exp,3)
width_exp = np.sqrt(np.var(data_in[ind_B_st:ind_B_end,1]))

# Set up an iterator for experimental data, normalize the data with respect to width_exp and stack plots vertically
def data_exp():
    for count,i in enumerate(files):
        data = np.loadtxt(i,skiprows=2)
        yield (data[:,0],(data[:,1]/width_exp+count*shift)/N_exp) 
data_exp = list(data_exp())

# load out-of-plane field simulation data and compute the variance of the 5 degree tilt data from 1 to 3 Tesla for normalization later
da_th = xr.load_dataarray('../../data/theory/OOP_Results/2020-11-22-13-43-31/oop_data.nc')[:,:,2]
B = da_th.coords['B']
N_th = da_th.coords['angle'].size
ind_B_st = index_find(B,1)
ind_B_end = index_find(B,3)
width_th = np.sqrt(np.var(da_th[0,ind_B_st:ind_B_end])).values
step = shift/N_exp

# load an in-plane simulation data and normalize it with the variacne from 1 to 3 Tesla
file_0 = '../../data/theory/in-plane/Pd_F_0.22_TAU_3.5_L_5.425.npy'
d_0tilt_th = np.load(file_0)
B_0tilt = d_0tilt_th[0,:]
ind_B_st = index_find(B_0tilt,1)
ind_B_end = index_find(B_0tilt,3)
d0_y = np.gradient(np.gradient(d_0tilt_th[1,:]))/d_0tilt_th[1,:]
d0_y =  d0_y/np.sqrt(np.var(d0_y[ind_B_st:ind_B_end]))

# Critical field for synchotron orbit depending on the angle 
def B_c(angle,width = 3.9):
    # can change the k_F
    C =  1e6*2*0.915e10*constants.hbar/constants.elementary_charge
    return np.divide(C/width,np.sin(angle*(np.pi/180)))
value_critical = np.linspace(0,0.9,100)
angle_critical = np.linspace(45,0.001,100)

#set up the color map and plot the results
cmap = plt.cm.viridis(np.linspace(0,1,N_exp))
linewidth = 5.90666
fig,ax1 = plt.subplots(1,1,figsize=(linewidth, linewidth*1.61803398875))
for count,i in enumerate(np.flip(np.arange(N_th))):
    index = index_find(B,B_c(45-5*count,width=5.425))
    ax1.plot(B[:index],(da_th[i,:index]/width_th+count*shift)/N_exp,color=cmap[count], ls='--')
ax1.plot(B_0tilt,(d0_y+shift*(N_exp-1))/N_exp,color=cmap[N_exp-1], ls='--')
ax1.set_xlim([0,10])
ax1.set_ylim([-0.05,1.2])
ax1.set_ylabel('$(d^2 \\rho/dB^2)/ \\rho$',fontsize=14)
ax1.set_xlabel('$ B[T] $',fontsize=14)
ax1.plot(B_c(angle_critical,width=5.45),value_critical,ls='--',c='k',label='Semiclassical Theory')
ax1.plot(B_c(angle_critical,width=3.9),value_critical,lw=2,c='k', label='Experiment')
ax1.annotate('$0^\circ$',
            xy=(10.4, 0.90), xycoords='data',annotation_clip=False)
ax1.annotate('$5^\circ$',
            xy=(10.4, 0.80), xycoords='data',annotation_clip=False)
ax1.annotate('$45^\circ$',
            xy=(10.4, 0), xycoords='data',annotation_clip=False)
ax1.annotate("",
            xy=(10.5, 0.05), xycoords='data',
            xytext=(10.5, 0.79), textcoords='data',
            arrowprops=dict(arrowstyle="->", connectionstyle="arc3", lw = 0.8),
            annotation_clip=False
            )
ax1.legend()
for count,i in enumerate(data_exp):
    ax1.plot(*i,color=cmap[count],lw=2)
    
fig.savefig("../figure4.svg",bbox_inches='tight',transparent=False)
