from scipy.interpolate import UnivariateSpline
from scipy.optimize import root_scalar
from scipy.integrate import solve_ivp, tplquad, quad, dblquad
from scipy.constants import constants
from autograd import grad
import autograd.numpy as np
import math
from numba import jit
from numpy import pi
### Setting up the natural unit system ###
#  distance is defined as: L_B = np.sqrt(HBAR/(EC) = 1
# time is in picoseconds
L_B_SI = math.sqrt(constants.hbar/(constants.elementary_charge))
C = 17.743/3*1e-10/L_B_SI

def to_cylindrical(func):
    def wrapper(r, theta, z, *args, **kwargs):
        return func(r*np.cos(theta), r*np.sin(theta), z, *args, **kwargs)
    return wrapper

def rotate(func, FROTATE):
    def wrapper(kxr, kyr, kzr, *args,**kwargs):
        return func(kxr*math.cos(FROTATE)-kyr*math.sin(FROTATE), kxr*math.sin(FROTATE)+kyr*math.cos(FROTATE), kzr, *args,**kwargs)
    return wrapper

def scale(func, scale_by):
    def wrapper(*args,**kwargs):
        return func(*args,**kwargs)*scale_by
    return wrapper


class Quasi2D():
    def __init__(self, energy, FS_ROTATE=np.pi/6, N=10000, flevel=0.22):
        """ A class that performs operations for quasi-2D materials

        Args:
            energy (function): An energy dispersion function in eV. Arguments are (kx,ky,kz) in Angstroms.
            FS_ROTATE(float, optional): Angle to rotate the Fermi surface.
            N (int, optional): Number of points used in the interpolation. Defaults to 10000.
            flevel (float, optional): The Fermi level in eV. Defaults to 0.22.
        """
        self._energy = rotate(energy, FS_ROTATE)
        self.energy_cl = to_cylindrical(self._energy)

        # Calculate velocity in micrometers/picoseconds
        self._vx_fn = scale(grad(self._energy, 0), 1/6.582119569)
        self._vy_fn = scale(grad(self._energy, 1), 1/6.582119569)
        self._vr_fn = scale(grad(self.energy_cl, 0), 1/6.582119569)


        # Generate velocity and wavevector data at fermi level in the natural unit system
        theta = np.linspace(-pi/2-0.1, 3*pi/2+0.1, N)
        kf = np.array([self.kf_value(i, 0, flevel) for i in theta])
        vx = np.array([self._vx_fn(kf[index]*np.cos(i), kf[index]*np.sin(i), 0) for index,i in enumerate(theta)])*1e-6/L_B_SI
        vy = np.array([self._vy_fn(kf[index]*np.cos(i), kf[index]*np.sin(i), 0) for index,i in enumerate(theta)])*1e-6/L_B_SI
        vr = np.array([self._vr_fn(kf[index], i, 0) for index, i in enumerate(theta)])*1e-6/L_B_SI
        kf *= 1.0e10*L_B_SI

        # Interpolate velocity and wavevector
        self._kf = UnivariateSpline(theta, kf, s=0)
        self._vr = UnivariateSpline(theta, vr, s=0)
        self._d_kf = UnivariateSpline(theta, np.gradient(kf, theta[1]-theta[0]), s=0)
        self.vx = UnivariateSpline(theta, vx, s=0)
        self.vy = UnivariateSpline(theta, vy, s=0)

    def kf_value(self, theta, kz, flevel, bounds=[0,1.5], x0=0.01):
        """ Generates the in-plane Fermi wavevector by bisecting the energy dispersion at the Fermi level

        Args:
            theta (float): the in-plane angle in the cylindrical k-space, in radians
            kz (float): kz value in cylindrical k-space, in Angstroms
            flevel (float): Fermi level, in eV
            bounds (list, optional): in-plane Fermi wavevector bounds for interpolation. Defaults to [0,1.5], in Angstroms.
            x0 (float, optional): initial guess. Defaults to 0.01, in Angstroms

        Returns:
            float: in-plane Fermi wavevector
        """
        _bisection_fn = lambda kf : self.energy_cl(kf,theta,kz)-flevel
        return root_scalar(_bisection_fn, bracket=bounds, x0=x0, rtol=1e-8, xtol=1e-14).root

    ### Trajectory Generation ###

    def trajectory_generation(self):
        """ Generates parametrized trajectories
        """

        # generate the ODE for theta(x)
        def f(tr, theta):
            return self.vx(theta)/(np.sin(theta)*self._d_kf(theta)+np.cos(theta)*self._kf(theta))

        # Establish the longest possible path along x direction to set up the the bounds
        self.tr_theta = solve_ivp((lambda theta,tr : 1/f(tr,theta)),[-pi/2,3*pi/2],[0],dense_output=True, method ='DOP853').sol
        self.TR_FULL = self.tr_theta(3*pi/2)[0]
        self.TR_SW = self.tr_theta(pi/2)[0]

        # theta(t)
        self.theta_tr = solve_ivp(f,[0,self.TR_FULL],[-pi/2],dense_output=True, method = 'DOP853').sol

        # t(x) whereas both time and position are in reduced units (t_actual = t_red * t_units)
        def vel(tr,xr):
            return self.vx(self.theta_tr(tr)[0])
        self.xr_tr = solve_ivp(vel,[0,self.TR_FULL],[0],dense_output=True, method = 'DOP853').sol
        self.XR_MAX = self.xr_tr(self.TR_SW)[0]

        # Establish the time when it hits the boundary
        self.tr_xr = solve_ivp((lambda xr,tr : 1/vel(tr,xr)),[0.0,self.XR_MAX],[0],dense_output=True, method = 'DOP853').sol

    # Equation of motion
    def eom(self, theta):
        return self.vx(theta)/(np.sin(theta)*self._d_kf(theta)+np.cos(theta)*self._kf(theta))

    # Calculate the current when the magnetic field is applied out-of-plane    
    def current_outplane(self,By,Bz,X_SAMPLE=3.9,TAU=5, error = 1e-04):
        """Given the out-of-plane magnetic field, calculates the magnetocurrent.

        Args:
            By (float): in-plane magnetic field, in Tesla
            Bz (float): out-of-plane magnetic field, in Tesla
            X_SAMPLE (float, optional): The width of the sample in micrometers.. Defaults to 3.9.
            TAU (int, optional): Scattering time in picoseconds. Defaults to 5.
            error ([float], optional): The relative error of the calculation. Defaults to 1e-04.

        Returns:
            float: magnetocurrent
        """
        TAU = TAU*Bz
        FREQ = C*By/Bz
        def integrand(trd,tr,theta0):
            tr_0 = self.tr_theta(theta0)[0]
            xr_0 = self.xr_tr(tr_0)[0]
            xrd = self.xr_tr(trd+tr_0)[0]-xr_0
            xr = self.xr_tr(tr+tr_0)[0]-xr_0

            exp = np.exp(-(tr-trd)/TAU)
            phase = FREQ*xr
            phase_d = FREQ*xrd
            # jacobian
            theta = self.theta_tr(tr_0+tr)[0]
            d_theta = self.eom(theta)
            d_theta0 = self.eom(theta0)
            jacobian =  d_theta * (1.0/d_theta0) * self.vx(theta0)

            return exp*(math.sin(phase)*math.sin(phase_d)+math.cos(phase)*math.cos(phase_d))*jacobian*(self._kf(theta)/self._vr(theta))

        # Edge t_limit
        def t_e_limit(theta0):
            return self.tr_theta(pi-theta0)[0]-self.tr_theta(theta0)[0]

        # Oscillation t_limit
        XR_SAMPLE = Bz*X_SAMPLE*1e-6/L_B_SI
        def t_o_limit(theta0):
            tr_0 = self.tr_theta(theta0)[0]
            xr_0 = self.xr_tr(tr_0)[0]
            return self.tr_xr(xr_0+XR_SAMPLE)[0]-tr_0

        XR_D = self.XR_MAX - XR_SAMPLE
        if XR_D <= 0:
            # Pure Edge current
            return np.array(tplquad(integrand, -pi/2, pi/2, lambda theta0: 0, t_e_limit, lambda theta0, tr: 0, lambda theta0, tr: tr, epsabs=0, epsrel=error))/Bz**2
        else:
            # Mixed current
            THETA_MAX = self.theta_tr(self.tr_xr(XR_D)[0])[0]
            o_current = np.array(tplquad(integrand, -pi/2, THETA_MAX, lambda theta0: 0, t_o_limit, lambda theta0, tr: 0, lambda theta0, tr: tr,epsabs=0, epsrel=error))
            e_current = np.array(tplquad(integrand, THETA_MAX, pi/2, lambda theta0: 0, t_e_limit, lambda theta0, tr: 0, lambda theta0, tr: tr, epsabs=0, epsrel=error))
            return (o_current+e_current)/Bz**2

    # Calculate current when the magnetic field is in-plane
    def current_inplane(self, B_array, X_SAMPLE, TAU):
        """Given in-plane magnetic field, calculates the magnetocurrent

        Args:
            B_array (list): in-plane magnetic field values for which to calculate the magnetocurrent, in Tesla
            X_SAMPLE (float): width of the sample, in micrometers.
            TAU (float): scattering time, in picoseconds
        Returns:
            float: magnetocurrent
        """

        # Convert to the natural unit system
        X_SAMPLE *= 1e-6/L_B_SI
    
        # Function that returns current per unit angle
        def I_theta_func(B):
            def I_theta(theta):
                l = TAU*self.vx(theta)
                r = l/X_SAMPLE
                phi = l*C
                omega = B*C*X_SAMPLE

                total = 1/(((B*phi)**2+1)**2)*((B*phi)**2*r + (B*phi)**2 + 1 - r + r*math.exp(-1/r)*(math.cos(omega)-2*B*phi*math.sin(omega)-(B*phi)**2*math.cos(omega)))

                return total*abs(self._kf(theta)/self._vr(theta))
            return I_theta

        # Integrate out all angle contributions to magnetoconductivity
        def Integrator(B_value):
            f = I_theta_func(B_value)
            return quad(f, -pi/2, pi/2)[0]

        # Vectorizing the integrator
        vecIntegrator = np.vectorize(Integrator)

        conduc = vecIntegrator(B_array)
        # Divide by the first value to normalize it
        conduc = np.divide(conduc, conduc[0])
        return conduc


    def current_2D(self, By, X_SAMPLE=3.9, Y_SAMPLE=5, error=1e-06):
        """Given in-plane magnetic field, calculates the magnetocurrent for realistic sample geometry. The calculation neglects relaxation.

        Args:
            B_array (list): in-plane magnetic field values for which to calculate the magnetocurrent, in Tesla
            X_SAMPLE (float): width of the sample, in micrometers.
            Y_SAMPLE (float): length of the sample, in micrometers.
            error(float, optional): relative error tolerance used in dblquad

        Returns:
            float: magnetocurrent
        """
        XR_SAMPLE = X_SAMPLE*1e-6/L_B_SI
        YR_SAMPLE = Y_SAMPLE*1e-6/L_B_SI

        @jit(nopython=True)
        def theta_1(x0):
            return math.atan(YR_SAMPLE/(XR_SAMPLE-x0))

        @jit(nopython=True)
        def theta_2(x0):
            return math.pi-math.atan(YR_SAMPLE/x0)

        @jit(nopython=True)
        def xb(y0, theta_0):

            theta_m = -1*math.atan(y0/XR_SAMPLE)
            theta_p = math.atan((YR_SAMPLE-y0)/XR_SAMPLE)

            if theta_0 > theta_p:
                return (YR_SAMPLE-y0)/math.tan(theta_0)
            elif theta_0 < theta_m:
                return y0/math.tan(-1*theta_0)
            else:
                return XR_SAMPLE

        freq = By*YR_SAMPLE*C

        def integrand1(theta_0, x0):
            vx = self.vx(theta_0)
            vy = self.vy(theta_0)

            jacobian = (self._kf(theta_0)/self._vr(theta_0))*abs(vy)
            denominator = By**2*vx**2
            vel_ratio = vx/vy
            return jacobian/denominator*(1-math.cos(freq*vel_ratio))

        def integrand2(y0, theta_0):
            vx = self.vx(theta_0)
            vy = self.vy(theta_0)

            jacobian = (self._kf(theta_0)/self._vr(theta_0))*abs(vx)
            denominator = By**2*vx**2
            return jacobian/denominator*(1-math.cos(By*C*xb(y0, theta_0)))

        result1 = dblquad(integrand1, 0, XR_SAMPLE, theta_1, theta_2, epsabs=0, epsrel=error)
        result2 = dblquad(integrand2, -math.pi/2, math.pi/2, 0, YR_SAMPLE, epsabs=0, epsrel=error)
        return [sum(x) for x in zip(result1,result2)]
